import hashlib
import hmac
import requests
import json
from .config import API_URL, API_VERSION, API_METHODS
from .utils import verify_request
from requests.exceptions import HTTPError


class Py3CW:

    def __init__(self, key: str, secret: str):
        """
        Init the library with a 3commas key and secret strings. Get keys from your account API
        (https://3commas.io/api_access_tokens) page
        """

        if key is None or key == '':
            raise ValueError('Please enter a 3commas API key')
        if secret is None or secret == '':
            raise ValueError('Please enter a 3commas API secret')

        self.key = key
        self.secret = secret

    def __generate_signature(self, path: str, data: str) -> str:
        """
        Generates the signature needed for 3commas API communication
        """
        encoded_key = str.encode(self.secret)
        message = str.encode(API_VERSION + path + data)
        signature = hmac.new(encoded_key, message, hashlib.sha256).hexdigest()
        return signature

    def __make_request(self, http_method: str, path: str, params: any, payload: any):
        """
        Private method that makes the actual request. Returns the response in JSON format for both
        success and error responses.
        """

        signature = self.__generate_signature(path, params)

        try:
            response = requests.request(
                method=http_method,
                url=API_URL + API_VERSION + path + '?' + params,
                headers={
                    'APIKEY': self.key,
                    'Signature': signature
                },
                data=payload
            )

            response_json = json.loads(response.text)

            if type(response_json) is dict and 'error' in response_json:
                return response_json, None
            else:
                return None, response_json

        except HTTPError as http_err:
            return {'error': True, 'msg': 'HTTP error occurred: {0}'.format(http_err)}, None

        except Exception as err:
            return {'error': True, 'msg': 'Other error occurred: {0}'.format(err)}, None

    @verify_request
    def request(self, entity: str, action: str = '', action_id: str = None, payload: any = None):
        """
        Constructs the API Url and makes the request.
        """

        api = API_METHODS[entity][action]
        method, api_path = api
        api_path = api_path.replace('{id}', action_id or '')

        return self.__make_request(
            http_method=method,
            path='{entity}{separator}{api_path}'.format(
                entity=entity,
                separator='/' if api_path else '',
                api_path=api_path or ''
            ),
            params='',
            payload=payload
        )



API_URL = 'https://api.3commas.io'
API_VERSION = '/public/api/ver1/'

# API methods from
# https://github.com/3commas-io/3commas-official-api-docs
API_METHODS = {
    'accounts': {
        '': ('GET', ''),
        'new': ('POST', 'new'),
        'update': ('POST', 'update'),
        'market_list': ('GET', 'market_list'),
        'market_pairs': ('GET', 'market_pairs'),
        'currency_rates': ('GET', 'currency_rates'),
        'active_trading_entities': ('GET', '{id}/active_trading_entities'),
        'sell_all_to_usd': ('POST', '{id}/sell_all_to_usd'),
        'sell_all_to_btc': ('POST', '{id}/sell_all_to_btc'),
        'balance_chart_data': ('GET', '{id}/balance_chart_data'),
        'load_balances': ('POST', '{id}/load_balances'),
        'rename': ('POST', '{id}/rename'),
        'pie_chart_data': ('POST', '{id}/pie_chart_data'),
        'account_table_data': ('POST', '{id}/account_table_data'),
        'remove': ('POST', '{id}/remove')
    },
    'deals': {
        '': ('GET', ''),
        'update_max_safety_orders': ('POST', '{id}/update_max_safety_orders'),
        'panic_sell': ('POST', '{id}/panic_sell'),
        'cancel': ('POST', '{id}/cancel'),
        'update_deal': ('PATCH', '{id}/update_deal'),
        'update_tp': ('POST', '{id}/update_tp'),
        'show': ('GET', '{id}/show'),
        'cancel_order': ('POST', '{id}/cancel_order'),
        'market_orders': ('GET', '{id}/market_orders'),
        'add_funds': ('POST', '{id}/add_funds'),
        'data_for_adding_funds': ('GET', '{id}/data_for_adding_funds'),
    },
    'bots': {
        '': ('GET', ''),
        'strategy_list': ('GET', 'strategy_list'),
        'pairs_black_list': ('GET', 'pairs_black_list'),
        'update_pairs_black_list': ('POST', 'update_pairs_black_list'),
        'create_bot': ('POST', 'create_bot'),
        'stats': ('GET', 'stats'),
        'update': ('PATCH', '{id}/update'),
        'disable': ('POST', '{id}/disable'),
        'enable': ('POST', '{id}/enable'),
        'start_new_deal': ('POST', '{id}/start_new_deal'),
        'delete': ('POST', '{id}/delete'),
        'panic_sell_all_deals': ('POST', '{id}/panic_sell_all_deals'),
        'cancel_all_deals': ('POST', '{id}/cancel_all_deals'),
        'show': ('GET', '{id}/show')
    },
    'grid_bots': {
        '': ('GET', ''),
        'ai': ('POST', 'ai'),
        'manual': ('POST', 'manual'),
        'ai_settings': ('GET', 'ai_settings'),
        'profits': ('GET', '{id}/profits'),
        'bot': ('GET', '{id}'),
        'delete_bot': ('DELETE', '{id}'),
        'disable_bot': ('POST', '{id}/disable'),
        'enable_bot': ('POST', '{id}/enable')
    },
    'marketplace': {
        'items': ('GET', 'items'),
        'signals': ('GET', '{id}/signals')
    },
    'smart_trades': {
        '': ('GET', ''),
        'create_simple_sell': ('POST', 'create_simple_sell'),
        'create_simple_buy': ('POST', 'create_simple_buy'),
        'create_smart_sell': ('POST', 'create_smart_sell'),
        'create_smart_cover': ('POST', 'create_smart_cover'),
        'create_smart_trade': ('POST', 'create_smart_trade'),
        'cancel_order': ('POST', '{id}/cancel_order'),
        'add_funds': ('POST', '{id}/add_funds'),
        'step_panic_sell': ('POST', '{id}/step_panic_sell'),
        'update': ('PATCH', '{id}/update'),
        'cancel': ('POST', '{id}/cancel'),
        'panic_sell': ('POST', '{id}/panic_sell'),
        'force_process': ('POST', '{id}/force_process'),
    }
}



from .config import API_METHODS


def verify_request(func):
    def wrapper(*args, **kw):
        """
        Raises an error if on of the arguments is missing or if the specific action requires and ID
        and it is not sent.
        """

        entity = kw.get('entity')
        action = kw.get('action')
        action_id = kw.get('action_id')
        api = API_METHODS[entity][action]
        method, api_path = api

        if entity is None or entity == '':
            raise ValueError('Missing entity')
        if entity not in API_METHODS:
            raise ValueError('Invalid entity')
        if action not in API_METHODS[entity]:
            raise ValueError('Invalid action')
        if '{id}' in api_path and (action_id is None or action_id == ''):
            raise ValueError('Missing ID for {action}'.format(action=action))

        return func(*args, **kw)

    return wrapper
